import { Schema as NormalizrSchema } from 'normalizr';
import { Record, Records, NormalizedData } from './Data';
import { Fields, Attribute } from './Attributes';
import { Attr } from './Attributes/Types';
import { HasOne, BelongsTo, HasMany, HasManyBy } from './Attributes/Relations';
export default class Model {
    /**
     * Name of the connection that this model is registerd.
     */
    static connection: string;
    /**
     * The name that is going be used as module name in Vuex Store.
     */
    static entity: string;
    /**
     * The primary key to be used for the model.
     */
    static primaryKey: string;
    [key: string]: any;
    /**
     * Creates a model instance.
     */
    constructor(data?: Record);
    /**
     * The definition of the fields of the model and its relations.
     */
    static fields(): Fields;
    /**
     * The generic attribute. The given value will be used as default value
     * of the property when instantiating a model.
     */
    static attr(value: any, mutator?: (value: any) => any): Attr;
    /**
     * Creates has one relationship.
     */
    static hasOne(model: typeof Model | string, foreignKey: string): HasOne;
    /**
     * Creates belongs to relationship.
     */
    static belongsTo(model: typeof Model | string, foreignKey: string): BelongsTo;
    /**
     * Creates has many relationship.
     */
    static hasMany(model: typeof Model | string, foreignKey: string): HasMany;
    /**
     * The has many by relationship.
     */
    static hasManyBy(model: typeof Model | string, foreignKey: string, otherKey: string): HasManyBy;
    /**
     * Mutators to mutate matching fields when instantiating the model.
     */
    static mutators(): {
        [field: string]: (value: any) => any;
    };
    /**
     * Get a model from the container.
     */
    static relation(name: string): typeof Model;
    /**
     * Resolve relation in the given attribute out of the container.
     */
    static resolveRelation(attr: HasOne | BelongsTo | HasMany | HasManyBy): typeof Model;
    /**
     * Create normalizr schema that represents this model.
     *
     * @param {boolean} many If true, it'll return an array schema.
     */
    static schema(many?: boolean): NormalizrSchema;
    /**
     * Generate normalized data from given data.
     */
    static normalize(data: any | any[]): NormalizedData;
    /**
     * Check if the given record has the appropriate foreign key and
     * if not, attach them.
     */
    static attachForeignKeys(records: Records, model: typeof Model): Records;
    /**
     * Get the static class of this model.
     */
    $self(): typeof Model;
    /**
     * Get the value of the primary key.
     */
    $id(): any;
    /**
     * The definition of the fields of the model and its relations.
     */
    $fields(): Fields;
    /**
     * Initialize the model by attaching all of the fields to property.
     */
    $initialize(data?: Record): void;
    /**
     * Merge given data into field's default value.
     */
    $merge(data?: Record): Fields;
    /**
     * Merge given data with fields and create a new fields.
     */
    $mergeFields(fields: Fields, data?: Record): Fields;
    /**
     * Build model by initializing given data.
     */
    $build(self: any, data: Fields): void;
    /**
     * Generate appropreate field value for the given attribute.
     */
    $generateField(attr: Attribute, key: string): any;
    /**
     * Resolve relation out of the container.
     */
    $resolveRelation(attr: HasOne | BelongsTo | HasMany | HasManyBy): typeof Model;
    /**
     * Serialize field values into json.
     */
    $toJson(): any;
    /**
     * Build Json data.
     */
    $buildJson(data: Fields, field: {
        [key: string]: any;
    }): any;
}
