"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("./support/lodash");
var Container_1 = require("./connections/Container");
var Data_1 = require("./Data");
var Schema_1 = require("./Schema");
var Attributes_1 = require("./Attributes");
var AttrTypes_1 = require("./Attributes/AttrTypes");
var Model = /** @class */ (function () {
    /**
     * Creates a model instance.
     */
    function Model(data) {
        this.$initialize(data);
    }
    /**
     * The definition of the fields of the model and its relations.
     */
    Model.fields = function () {
        return {};
    };
    /**
     * The generic attribute. The given value will be used as default value
     * of the property when instantiating a model.
     */
    Model.attr = function (value, mutator) {
        return Attributes_1.default.attr(value, mutator);
    };
    /**
     * Creates has one relationship.
     */
    Model.hasOne = function (model, foreignKey) {
        return Attributes_1.default.hasOne(model, foreignKey);
    };
    /**
     * Creates belongs to relationship.
     */
    Model.belongsTo = function (model, foreignKey) {
        return Attributes_1.default.belongsTo(model, foreignKey);
    };
    /**
     * Creates has many relationship.
     */
    Model.hasMany = function (model, foreignKey) {
        return Attributes_1.default.hasMany(model, foreignKey);
    };
    /**
     * The has many by relationship.
     */
    Model.hasManyBy = function (model, foreignKey, otherKey) {
        return Attributes_1.default.hasManyBy(model, foreignKey, otherKey);
    };
    /**
     * Mutators to mutate matching fields when instantiating the model.
     */
    Model.mutators = function () {
        return {};
    };
    /**
     * Get a model from the container.
     */
    Model.relation = function (name) {
        return Container_1.default.connection(this.connection).model(name);
    };
    /**
     * Resolve relation in the given attribute out of the container.
     */
    Model.resolveRelation = function (attr) {
        return _.isString(attr.model) ? this.relation(attr.model) : attr.model;
    };
    /**
     * Create normalizr schema that represents this model.
     *
     * @param {boolean} many If true, it'll return an array schema.
     */
    Model.schema = function (many) {
        if (many === void 0) { many = false; }
        return many ? Schema_1.default.many(this) : Schema_1.default.one(this);
    };
    /**
     * Generate normalized data from given data.
     */
    Model.normalize = function (data) {
        var _this = this;
        var schema = this.schema(_.isArray(data));
        var normalizedData = Data_1.default.normalize(data, schema);
        // Check if all foreign keys exist in the data and if not, make them.
        return _.mapValues(normalizedData, function (records, entity) {
            return _this.attachForeignKeys(records, _this.relation(entity));
        });
    };
    /**
     * Check if the given record has the appropriate foreign key and
     * if not, attach them.
     */
    Model.attachForeignKeys = function (records, model) {
        var fields = model.fields();
        return _.mapValues(records, function (record) {
            var newRecord = __assign({}, record);
            _.forEach(record, function (value, field) {
                var attr = fields[field];
                if (!attr) {
                    return;
                }
                if (!Attributes_1.default.isAttribute(attr)) {
                    return;
                }
                if (attr.type === AttrTypes_1.default.Attr) {
                    return;
                }
                if (attr.type === AttrTypes_1.default.BelongsTo) {
                    var key = attr.foreignKey;
                    if (newRecord[key]) {
                        return;
                    }
                    newRecord[key] = value;
                }
            });
            return newRecord;
        });
    };
    /**
     * Get the static class of this model.
     */
    Model.prototype.$self = function () {
        return this.constructor;
    };
    /**
     * Get the value of the primary key.
     */
    Model.prototype.$id = function () {
        return this[this.$self().primaryKey];
    };
    /**
     * The definition of the fields of the model and its relations.
     */
    Model.prototype.$fields = function () {
        return this.$self().fields();
    };
    /**
     * Initialize the model by attaching all of the fields to property.
     */
    Model.prototype.$initialize = function (data) {
        var fields = this.$merge(data);
        this.$build(this, fields);
    };
    /**
     * Merge given data into field's default value.
     */
    Model.prototype.$merge = function (data) {
        if (!data) {
            return this.$fields();
        }
        var fields = __assign({}, this.$fields());
        return this.$mergeFields(fields, data);
    };
    /**
     * Merge given data with fields and create a new fields.
     */
    Model.prototype.$mergeFields = function (fields, data) {
        var _this = this;
        var keys = _.keys(fields);
        _.forEach(data, function (value, key) {
            if (!_.includes(keys, key)) {
                return;
            }
            if (Attributes_1.default.isAttribute(fields[key])) {
                fields[key].value = value;
                return;
            }
            fields[key] = _this.$mergeFields(fields[key], value);
        });
        return fields;
    };
    /**
     * Build model by initializing given data.
     */
    Model.prototype.$build = function (self, data) {
        var _this = this;
        _.forEach(data, function (field, key) {
            if (Attributes_1.default.isAttribute(field)) {
                self[key] = _this.$generateField(field, key);
                return;
            }
            _this.$build(self[key] = {}, field);
        });
    };
    /**
     * Generate appropreate field value for the given attribute.
     */
    Model.prototype.$generateField = function (attr, key) {
        if (attr.value === null) {
            return null;
        }
        if (attr.type === AttrTypes_1.default.Attr) {
            var mutator = attr.mutator || this.$self().mutators()[key];
            return mutator ? mutator(attr.value) : attr.value;
        }
        if (_.isNumber(attr.value) || _.isNumber(attr.value[0])) {
            return null;
        }
        var model = this.$resolveRelation(attr);
        if (attr.type === AttrTypes_1.default.HasOne || attr.type === AttrTypes_1.default.BelongsTo) {
            return attr.value ? new model(attr.value) : null;
        }
        if (attr.type === AttrTypes_1.default.HasMany || attr.type === AttrTypes_1.default.HasManyBy) {
            return attr.value ? attr.value.map(function (v) { return new model(v); }) : null;
        }
    };
    /**
     * Resolve relation out of the container.
     */
    Model.prototype.$resolveRelation = function (attr) {
        return this.$self().resolveRelation(attr);
    };
    /**
     * Serialize field values into json.
     */
    Model.prototype.$toJson = function () {
        return this.$buildJson(this.$self().fields(), this);
    };
    /**
     * Build Json data.
     */
    Model.prototype.$buildJson = function (data, field) {
        return _.mapValues(data, function (attr, key) {
            if (!field[key]) {
                return field[key];
            }
            if (!Attributes_1.default.isAttribute(attr)) {
                return field.$buildJson(attr, field[key]);
            }
            if (attr.type === AttrTypes_1.default.HasOne || attr.type === AttrTypes_1.default.BelongsTo) {
                return field[key].$toJson();
            }
            if (attr.type === AttrTypes_1.default.HasMany) {
                return field[key].map(function (model) { return model.$toJson(); });
            }
            return field[key];
        });
    };
    /**
     * The primary key to be used for the model.
     */
    Model.primaryKey = 'id';
    return Model;
}());
exports.default = Model;
//# sourceMappingURL=Model.js.map