import { Record } from './Data';
import { State, EntityState } from './Module';
export declare type Item = Record | null;
export declare type Collection = Record[];
export declare type WhereBoolean = 'and' | 'or';
export declare type OrderDirection = 'asc' | 'desc';
export declare type Predicate = (item: Record) => boolean;
export declare type Condition = number | string | Predicate;
export interface Wheres {
    field: any;
    value: any;
    boolean: WhereBoolean;
}
export interface Orders {
    field: string;
    direction: OrderDirection;
}
export default class Query {
    /**
     * The Vuex Store State. This is the target where query will perform
     * CRUD actions.
     */
    protected state: State;
    /**
     * The name of the entity.
     */
    protected name: string;
    /**
     * The primary key of the entity.
     */
    protected primaryKey: string;
    /**
     * The Vuex Store State for of the entity.
     */
    protected entity: EntityState;
    /**
     * The records that have been processed.
     */
    protected records: Collection;
    /**
     * The where constraints for the query.
     */
    protected wheres: Wheres[];
    /**
     * The orders of the query result.
     */
    protected orders: Orders[];
    /**
     * Number of results to skip.
     */
    protected _offset: number;
    /**
     * Maximum number of records to return.
     *
     * We use polyfill of `Number.MAX_SAFE_INTEGER` for IE11 here.
     */
    protected _limit: number;
    /**
     * Create a new query instance.
     */
    constructor(state: State, name: string);
    /**
     * Returns single record of the query chain result.
     */
    get(): Collection;
    /**
     * Returns single record of the query chain result.
     */
    first(id?: number | string): Item;
    /**
     * Process the query and filter data.
     */
    process(): void;
    /**
     * Add a and where clause to the query.
     */
    where(field: any, value?: any): this;
    /**
     * Save the given data to the state. This will replace any existing
     * data in the state.
     */
    create(data: any): void;
    /**
     * Insert given data to the state. Unlike `create`, this method will not
     * remove existing data within the state, but it will update the data
     * with the same primary key.
     */
    insert(data: any): void;
    /**
     * Update data in the state.
     */
    update(data: Record, condition: Condition): void;
    /**
     * Add a or where clause to the query.
     */
    orWhere(field: any, value?: any): this;
    /**
     * Add an order to the query.
     */
    orderBy(field: string, direction?: OrderDirection): this;
    /**
     * Add an offset to the query.
     */
    offset(offset: number): this;
    /**
     * Add limit to the query.
     */
    limit(limit: number): this;
    /**
     * Delete data from the state.
     */
    delete(condition: Condition): void;
    /**
     * Delete all data from the state.
     */
    deleteAll(): void;
    /**
     * Create a item from given record.
     */
    item(record?: Record | null): Item;
    /**
     * Create a collection (array) from given records.
     */
    collect(): Collection;
    /**
     * Filter the given data by registered where clause.
     */
    selectByWheres(): void;
    /**
     * Sort the given data by registered orders.
     */
    sortByOrders(): void;
    /**
     * Checks if given Record matches the registered where clause.
     */
    whereOnRecord(record: Record): boolean;
    /**
     * Get comparator for the where clause.
     */
    getComparator(record: Record): (where: any) => boolean;
}
