"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("./support/lodash");
var Query = /** @class */ (function () {
    /**
     * Create a new query instance.
     */
    function Query(state, name) {
        /**
         * The records that have been processed.
         */
        this.records = [];
        /**
         * The where constraints for the query.
         */
        this.wheres = [];
        /**
         * The orders of the query result.
         */
        this.orders = [];
        /**
         * Number of results to skip.
         */
        this._offset = 0;
        /**
         * Maximum number of records to return.
         *
         * We use polyfill of `Number.MAX_SAFE_INTEGER` for IE11 here.
         */
        this._limit = Math.pow(2, 53) - 1;
        this.state = state;
        this.name = name;
        this.entity = state[name];
        this.primaryKey = state[name].$primaryKey || 'id';
        this.records = _.map(state[name].data, function (record) { return record; });
    }
    /**
     * Returns single record of the query chain result.
     */
    Query.prototype.get = function () {
        this.process();
        return this.collect();
    };
    /**
     * Returns single record of the query chain result.
     */
    Query.prototype.first = function (id) {
        this.process();
        if (_.isEmpty(this.records)) {
            return null;
        }
        if (id !== undefined) {
            return this.item(_.find(this.records, [this.primaryKey, id]));
        }
        return this.item(this.records[0]);
    };
    /**
     * Process the query and filter data.
     */
    Query.prototype.process = function () {
        // If the where clause is registered, lets filter the records beased on it.
        if (!_.isEmpty(this.wheres)) {
            this.selectByWheres();
        }
        // Next, lets sort the data if orderBy is registred.
        if (!_.isEmpty(this.orders)) {
            this.sortByOrders();
        }
        // Finally, slice the record by limit and offset.
        this.records = _.slice(this.records, this._offset, this._offset + this._limit);
    };
    /**
     * Add a and where clause to the query.
     */
    Query.prototype.where = function (field, value) {
        this.wheres.push({ field: field, value: value, boolean: 'and' });
        return this;
    };
    /**
     * Save the given data to the state. This will replace any existing
     * data in the state.
     */
    Query.prototype.create = function (data) {
        this.entity.data = data;
    };
    /**
     * Insert given data to the state. Unlike `create`, this method will not
     * remove existing data within the state, but it will update the data
     * with the same primary key.
     */
    Query.prototype.insert = function (data) {
        this.entity.data = __assign({}, this.entity.data, data);
    };
    /**
     * Update data in the state.
     */
    Query.prototype.update = function (data, condition) {
        if (typeof condition !== 'function') {
            if (this.entity.data[condition]) {
                this.entity.data[condition] = __assign({}, this.entity.data[condition], data);
            }
            return;
        }
        this.entity.data = _.mapValues(this.entity.data, function (record) {
            return condition(record) ? __assign({}, record, data) : record;
        });
    };
    /**
     * Add a or where clause to the query.
     */
    Query.prototype.orWhere = function (field, value) {
        this.wheres.push({ field: field, value: value, boolean: 'or' });
        return this;
    };
    /**
     * Add an order to the query.
     */
    Query.prototype.orderBy = function (field, direction) {
        if (direction === void 0) { direction = 'asc'; }
        this.orders.push({ field: field, direction: direction });
        return this;
    };
    /**
     * Add an offset to the query.
     */
    Query.prototype.offset = function (offset) {
        this._offset = offset;
        return this;
    };
    /**
     * Add limit to the query.
     */
    Query.prototype.limit = function (limit) {
        this._limit = limit;
        return this;
    };
    /**
     * Delete data from the state.
     */
    Query.prototype.delete = function (condition) {
        if (typeof condition === 'function') {
            this.entity.data = _.pickBy(this.entity.data, function (record) { return !condition(record); });
            return;
        }
        var id = typeof condition === 'number' ? condition.toString() : condition;
        this.entity.data = _.pickBy(this.entity.data, function (_record, key) { return key !== id; });
    };
    /**
     * Delete all data from the state.
     */
    Query.prototype.deleteAll = function () {
        this.entity.data = {};
    };
    /**
     * Create a item from given record.
     */
    Query.prototype.item = function (record) {
        return record ? record : null;
    };
    /**
     * Create a collection (array) from given records.
     */
    Query.prototype.collect = function () {
        return !_.isEmpty(this.records) ? this.records : [];
    };
    /**
     * Filter the given data by registered where clause.
     */
    Query.prototype.selectByWheres = function () {
        var _this = this;
        this.records = this.records.filter(function (record) { return _this.whereOnRecord(record); });
    };
    /**
     * Sort the given data by registered orders.
     */
    Query.prototype.sortByOrders = function () {
        var keys = _.map(this.orders, 'field');
        var directions = _.map(this.orders, 'direction');
        this.records = _.orderBy(this.records, keys, directions);
    };
    /**
     * Checks if given Record matches the registered where clause.
     */
    Query.prototype.whereOnRecord = function (record) {
        var whereTypes = _.groupBy(this.wheres, function (where) { return where.boolean; });
        var whereResults = [];
        var comparator = this.getComparator(record);
        if (whereTypes.and) {
            whereResults.push(_.every(whereTypes.and, comparator));
        }
        if (whereTypes.or) {
            whereResults.push(_.some(whereTypes.or, comparator));
        }
        return whereResults.indexOf(true) !== -1;
    };
    /**
     * Get comparator for the where clause.
     */
    Query.prototype.getComparator = function (record) {
        var _this = this;
        return function (where) {
            // Function with Record and Query as argument.
            if (_.isFunction(where.field)) {
                var query = new Query(_this.state, _this.name);
                var result = where.field(record, query);
                if (typeof result === 'boolean') {
                    return result;
                }
                return !_.isEmpty(query.where(_this.primaryKey, record[_this.primaryKey]).get());
            }
            // Function with Record value as argument.
            if (_.isFunction(where.value)) {
                return where.value(record[where.field]);
            }
            // Check if field value is in given where Array.
            if (_.isArray(where.value)) {
                return where.value.indexOf(record[where.field]) !== -1;
            }
            // Simple equal check.
            return record[where.field] === where.value;
        };
    };
    return Query;
}());
exports.default = Query;
//# sourceMappingURL=Query.js.map