import { Record, NormalizedData } from './Data';
import { Fields } from './Attributes';
import { HasOne, BelongsTo, HasMany, HasManyBy } from './Attributes/Relations';
import Model from './Model';
import { State } from './Module';
import Query, { Item as QueryItem, Collection as QueryCollection, OrderDirection, Condition } from './Query';
export declare type Item = Model | Record | null;
export declare type Collection = Model[] | Record[];
export declare type Buildable = QueryItem | QueryCollection | null;
export declare type Constraint = (query: Repo) => void | boolean;
export interface Relation {
    name: string;
    constraint: null | Constraint;
}
export default class Repo {
    /**
     * The base query builder instance.
     */
    protected query: Query;
    /**
     * The Vuex Store State.
     */
    protected state: State;
    /**
     * The name of the entity.
     */
    protected name: string;
    /**
     * The model of the entity.
     */
    protected entity: typeof Model;
    /**
     * Whether to wrap returing record with class or to return as plain object.
     */
    protected wrap: boolean;
    /**
     * The relationships that should be loaded with the result.
     */
    protected load: Relation[];
    /**
     * Create a new repo instance.
     */
    constructor(state: State, entity: string, wrap?: boolean);
    /**
     * Create a new repo instance
     */
    static query(state: State, name: string, wrap?: boolean): Repo;
    /**
     * Get all data of the given entity from the state.
     */
    static all(state: State, entity: string, wrap?: boolean): Collection;
    /**
     * Find a data of the given entity by given id from the given state.
     */
    static find(state: State, entity: string, id: string | number, wrap?: boolean): Item;
    /**
     * Get the count of the retrieved data.
     */
    static count(state: State, entity: string, wrap?: boolean): number;
    /**
     * Save the given data to the state. This will replace any existing
     * data in the state.
     */
    static create(state: State, entity: string, data: any): void;
    /**
     * Insert given data to the state. Unlike `create`, this method will not
     * remove existing data within the state, but it will update the data
     * with the same primary key.
     */
    static insert(state: State, entity: string, data: any): void;
    /**
     * Update data in the state.
     */
    static update(state: State, entity: string, data: any, condition?: Condition): void;
    /**
     * Delete data from the state.
     */
    static delete(state: State, entity: string, condition: Condition): void;
    /**
     * Delete all data from the state.
     */
    static deleteAll(state: State, entity?: string): void;
    /**
     * Get model of given name from connections container.
     */
    static model(state: State, name: string): typeof Model;
    /**
     * Get all models from connections container.
     */
    static models(state: State): {
        [name: string]: typeof Model;
    };
    /**
     * Get the primary key for the record.
     */
    static primaryKey(state: State, name: string): string;
    /**
     * Get Repo class.
     */
    self(): typeof Repo;
    /**
     * Get model of given name from connections container.
     */
    model(name: string): typeof Model;
    /**
     * Get all models from connections container.
     */
    models(): {
        [name: string]: typeof Model;
    };
    /**
     * Get the primary key of the model.
     */
    primaryKey(): string;
    /**
     * Returns all record of the query chain result. This method is alias
     * of the `get` method.
     */
    all(): Collection;
    /**
     * Returns single record of the query chain result. This method is alias
     * of the `first` method.
     */
    find(id: number | string): Item;
    /**
     * Returns all record of the query chain result.
     */
    get(): Collection;
    /**
     * Returns single record of the query chain result.
     */
    first(id?: number | string): Item;
    /**
     * Add a and where clause to the query.
     */
    where(field: any, value?: any): this;
    /**
     * Add a or where clause to the query.
     */
    orWhere(field: any, value?: any): this;
    /**
     * Add an order to the query.
     */
    orderBy(field: string, direction?: OrderDirection): this;
    /**
     * Add an offset to the query.
     */
    offset(offset: number): this;
    /**
     * Add limit to the query.
     */
    limit(limit: number): this;
    /**
     * Set the relationships that should be loaded.
     */
    with(name: string, constraint?: Constraint | null): this;
    /**
     * Set where constraint based on relationship existence.
     */
    has(name: string, constraint?: number | string | Constraint | null, count?: number): this;
    /**
     * Set where constraint based on relationship absence.
     */
    hasNot(name: string, constraint?: number | string | Constraint | null, count?: number): this;
    /**
     * Add where constraints based on has or hasNot condition.
     */
    addHasConstraint(name: string, constraint?: number | string | Constraint | null, count?: number, existence?: boolean): this;
    /**
     * Save the given data to the state. This will replace any existing
     * data in the state.
     */
    create(data: any): void;
    /**
     * Insert given data to the state. Unlike `create`, this method will not
     * remove existing data within the state, but it will update the data
     * with the same primary key.
     */
    insert(data: any): void;
    /**
     * Save data into Vuex Store.
     */
    save(method: string, data: any): void;
    /**
     * Get the count of the retrieved data.
     */
    count(): number;
    /**
     * Fill missing fields in given data with default value defined in
     * corresponding model.
     */
    fill(data: Record, entity: string): Record;
    /**
     * Build record.
     */
    buildRecord(data: any, fields: Fields, record?: Record): Record;
    /**
     * Update data in the state.
     */
    update(data: any, condition?: Condition): void;
    /**
     * Delete data from the state.
     */
    delete(condition: Condition): void;
    /**
     * Delete all data from the state.
     */
    deleteAll(): void;
    /**
     * Create a item from given record.
     */
    item(queryItem: QueryItem): Item;
    /**
     * Create a collection (array) from given records.
     */
    collect(collection: QueryCollection): Collection;
    /**
     * Load the relationships for the record.
     */
    loadRelations(base: Record, load?: Relation[], record?: Record, fields?: Fields): Record;
    /**
     * Load the has one relationship for the record.
     */
    loadHasOneRelation(record: Record, attr: HasOne, relation: Relation): Record | null;
    /**
     * Load the belongs to relationship for the record.
     */
    loadBelongsToRelation(record: Record, attr: BelongsTo, relation: Relation): Record | null;
    /**
     * Load the has many relationship for the record.
     */
    loadHasManyRelation(record: Record, attr: HasMany, relation: Relation): Record[] | null;
    /**
     * Load the has many by relationship for the record.
     */
    loadHasManyByRelation(record: Record, attr: HasManyBy, relation: Relation): Record[] | null;
    /**
     * Resolve relation out of the container.
     */
    resolveRelation(attr: HasOne | BelongsTo | HasMany | HasManyBy): typeof Model;
    /**
     * Check if the given record has given relationship.
     */
    hasRelation(record: Record, name: string, constraint?: number | string | Constraint | null, count?: number): boolean;
    /**
     * Add constraint to the query.
     */
    addConstraint(query: Repo, relation: Relation): void;
    /**
     * Normalize the given data by given model.
     */
    normalize(data: any): NormalizedData;
}
