"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("./support/lodash");
var Container_1 = require("./connections/Container");
var AttrTypes_1 = require("./Attributes/AttrTypes");
var Attributes_1 = require("./Attributes");
var Query_1 = require("./Query");
var Repo = /** @class */ (function () {
    /**
     * Create a new repo instance.
     */
    function Repo(state, entity, wrap) {
        if (wrap === void 0) { wrap = true; }
        /**
         * The relationships that should be loaded with the result.
         */
        this.load = [];
        this.state = state;
        this.name = entity;
        this.entity = this.model(entity);
        this.wrap = wrap;
        this.query = new Query_1.default(state, entity);
    }
    /**
     * Create a new repo instance
     */
    Repo.query = function (state, name, wrap) {
        if (wrap === void 0) { wrap = true; }
        return new this(state, name, wrap);
    };
    /**
     * Get all data of the given entity from the state.
     */
    Repo.all = function (state, entity, wrap) {
        if (wrap === void 0) { wrap = true; }
        return (new this(state, entity, wrap)).get();
    };
    /**
     * Find a data of the given entity by given id from the given state.
     */
    Repo.find = function (state, entity, id, wrap) {
        if (wrap === void 0) { wrap = true; }
        return (new this(state, entity, wrap)).first(id);
    };
    /**
     * Get the count of the retrieved data.
     */
    Repo.count = function (state, entity, wrap) {
        if (wrap === void 0) { wrap = false; }
        return (new this(state, entity, wrap)).count();
    };
    /**
     * Save the given data to the state. This will replace any existing
     * data in the state.
     */
    Repo.create = function (state, entity, data) {
        (new this(state, entity)).create(data);
    };
    /**
     * Insert given data to the state. Unlike `create`, this method will not
     * remove existing data within the state, but it will update the data
     * with the same primary key.
     */
    Repo.insert = function (state, entity, data) {
        (new this(state, entity)).insert(data);
    };
    /**
     * Update data in the state.
     */
    Repo.update = function (state, entity, data, condition) {
        (new this(state, entity)).update(data, condition);
    };
    /**
     * Delete data from the state.
     */
    Repo.delete = function (state, entity, condition) {
        (new this(state, entity)).delete(condition);
    };
    /**
     * Delete all data from the state.
     */
    Repo.deleteAll = function (state, entity) {
        if (!entity) {
            var models_1 = this.models(state);
            Object.keys(models_1).forEach(function (key) {
                var entityName = models_1[key].entity;
                if (state[entityName]) {
                    state[entityName].data = {};
                }
            });
            return;
        }
        (new this(state, entity)).deleteAll();
    };
    /**
     * Get model of given name from connections container.
     */
    Repo.model = function (state, name) {
        return Container_1.default.connection(state.name).model(name);
    };
    /**
     * Get all models from connections container.
     */
    Repo.models = function (state) {
        return Container_1.default.connection(state.name).models();
    };
    /**
     * Get the primary key for the record.
     */
    Repo.primaryKey = function (state, name) {
        return this.model(state, name).primaryKey;
    };
    /**
     * Get Repo class.
     */
    Repo.prototype.self = function () {
        return this.constructor;
    };
    /**
     * Get model of given name from connections container.
     */
    Repo.prototype.model = function (name) {
        return this.self().model(this.state, name);
    };
    /**
     * Get all models from connections container.
     */
    Repo.prototype.models = function () {
        return this.self().models(this.state);
    };
    /**
     * Get the primary key of the model.
     */
    Repo.prototype.primaryKey = function () {
        return this.self().primaryKey(this.state, this.name);
    };
    /**
     * Returns all record of the query chain result. This method is alias
     * of the `get` method.
     */
    Repo.prototype.all = function () {
        return this.get();
    };
    /**
     * Returns single record of the query chain result. This method is alias
     * of the `first` method.
     */
    Repo.prototype.find = function (id) {
        return this.first(id);
    };
    /**
     * Returns all record of the query chain result.
     */
    Repo.prototype.get = function () {
        return this.collect(this.query.get());
    };
    /**
     * Returns single record of the query chain result.
     */
    Repo.prototype.first = function (id) {
        return this.item(this.query.first(id));
    };
    /**
     * Add a and where clause to the query.
     */
    Repo.prototype.where = function (field, value) {
        this.query.where(field, value);
        return this;
    };
    /**
     * Add a or where clause to the query.
     */
    Repo.prototype.orWhere = function (field, value) {
        this.query.orWhere(field, value);
        return this;
    };
    /**
     * Add an order to the query.
     */
    Repo.prototype.orderBy = function (field, direction) {
        if (direction === void 0) { direction = 'asc'; }
        this.query.orderBy(field, direction);
        return this;
    };
    /**
     * Add an offset to the query.
     */
    Repo.prototype.offset = function (offset) {
        this.query.offset(offset);
        return this;
    };
    /**
     * Add limit to the query.
     */
    Repo.prototype.limit = function (limit) {
        this.query.limit(limit);
        return this;
    };
    /**
     * Set the relationships that should be loaded.
     */
    Repo.prototype.with = function (name, constraint) {
        if (constraint === void 0) { constraint = null; }
        this.load.push({ name: name, constraint: constraint });
        return this;
    };
    /**
     * Set where constraint based on relationship existence.
     */
    Repo.prototype.has = function (name, constraint, count) {
        if (constraint === void 0) { constraint = null; }
        return this.addHasConstraint(name, constraint, count, true);
    };
    /**
     * Set where constraint based on relationship absence.
     */
    Repo.prototype.hasNot = function (name, constraint, count) {
        if (constraint === void 0) { constraint = null; }
        return this.addHasConstraint(name, constraint, count, false);
    };
    /**
     * Add where constraints based on has or hasNot condition.
     */
    Repo.prototype.addHasConstraint = function (name, constraint, count, existence) {
        var _this = this;
        if (constraint === void 0) { constraint = null; }
        if (existence === void 0) { existence = true; }
        var id = this.primaryKey();
        var ids = [];
        var items = (new Query_1.default(this.state, this.name)).get();
        _.forEach(items, function (item) {
            _this.hasRelation(item, name, constraint, count) === existence && ids.push(item[id]);
        });
        this.where(id, function (key) { return _.includes(ids, key); });
        return this;
    };
    /**
     * Save the given data to the state. This will replace any existing
     * data in the state.
     */
    Repo.prototype.create = function (data) {
        this.save('create', data);
    };
    /**
     * Insert given data to the state. Unlike `create`, this method will not
     * remove existing data within the state, but it will update the data
     * with the same primary key.
     */
    Repo.prototype.insert = function (data) {
        this.save('insert', data);
    };
    /**
     * Save data into Vuex Store.
     */
    Repo.prototype.save = function (method, data) {
        var _this = this;
        var normalizedData = this.normalize(data);
        // Update with empty data.
        if (method === 'create' && _.isEmpty(normalizedData)) {
            this.query[method](normalizedData);
            return;
        }
        _.forEach(normalizedData, function (data, entity) {
            var filledData = _.mapValues(data, function (record) { return _this.fill(record, entity); });
            entity === _this.name ? _this.query[method](filledData) : new Query_1.default(_this.state, entity)[method](filledData);
        });
    };
    /**
     * Get the count of the retrieved data.
     */
    Repo.prototype.count = function () {
        // Do not wrap result data with class because it's unnecessary.
        this.wrap = false;
        return this.get().length;
    };
    /**
     * Fill missing fields in given data with default value defined in
     * corresponding model.
     */
    Repo.prototype.fill = function (data, entity) {
        return this.buildRecord(data, this.model(entity).fields());
    };
    /**
     * Build record.
     */
    Repo.prototype.buildRecord = function (data, fields, record) {
        var _this = this;
        if (record === void 0) { record = {}; }
        var newRecord = record;
        _.forEach(fields, function (attr, name) {
            if (Attributes_1.default.isFields(attr)) {
                var newData = data[name] ? data[name] : {};
                newRecord[name] = _this.buildRecord(newData, attr, newRecord[name]);
                return;
            }
            if (data[name] !== undefined) {
                newRecord[name] = data[name];
                return;
            }
            if (attr.type === AttrTypes_1.default.Attr) {
                newRecord[name] = attr.value;
                return;
            }
            if (attr.type === AttrTypes_1.default.HasOne || attr.type === AttrTypes_1.default.BelongsTo) {
                newRecord[name] = null;
                return;
            }
            if (attr.type === AttrTypes_1.default.HasMany || attr.type === AttrTypes_1.default.HasManyBy) {
                newRecord[name] = [];
                return;
            }
        });
        return newRecord;
    };
    /**
     * Update data in the state.
     */
    Repo.prototype.update = function (data, condition) {
        // If there is no condition, check if data contains primary key. If it has,
        // use it as a condition to update the data. Else, do nothing.
        if (!condition) {
            data[this.entity.primaryKey] && this.query.update(data, data[this.entity.primaryKey]);
            return;
        }
        this.query.update(data, condition);
    };
    /**
     * Delete data from the state.
     */
    Repo.prototype.delete = function (condition) {
        this.query.delete(condition);
    };
    /**
     * Delete all data from the state.
     */
    Repo.prototype.deleteAll = function () {
        this.query.deleteAll();
    };
    /**
     * Create a item from given record.
     */
    Repo.prototype.item = function (queryItem) {
        if (!queryItem) {
            return null;
        }
        var item = queryItem;
        if (!_.isEmpty(this.load)) {
            item = this.loadRelations(item);
        }
        if (!this.wrap) {
            return item;
        }
        return new this.entity(item);
    };
    /**
     * Create a collection (array) from given records.
     */
    Repo.prototype.collect = function (collection) {
        var _this = this;
        if (_.isEmpty(collection)) {
            return [];
        }
        var item = collection;
        if (!_.isEmpty(this.load)) {
            item = _.map(item, function (data) { return _this.loadRelations(data); });
        }
        if (!this.wrap) {
            return item;
        }
        return _.map(item, function (data) { return new _this.entity(data); });
    };
    /**
     * Load the relationships for the record.
     */
    Repo.prototype.loadRelations = function (base, load, record, fields) {
        var _this = this;
        var _load = load || this.load;
        var _record = record || __assign({}, base);
        var _fields = fields || this.entity.fields();
        return _.reduce(_load, function (record, relation) {
            var name = relation.name.split('.')[0];
            var attr = _fields[name];
            if (!attr || !Attributes_1.default.isAttribute(attr)) {
                _.forEach(_fields, function (f, key) {
                    if (f[name]) {
                        record[key] = _this.loadRelations(base, _load, record[key], f);
                        return;
                    }
                });
                return record;
            }
            if (attr.type === AttrTypes_1.default.Attr) {
                return record;
            }
            if (attr.type === AttrTypes_1.default.HasOne) {
                record[name] = _this.loadHasOneRelation(base, attr, relation);
                return record;
            }
            if (attr.type === AttrTypes_1.default.BelongsTo) {
                record[name] = _this.loadBelongsToRelation(base, attr, relation);
                return record;
            }
            if (attr.type === AttrTypes_1.default.HasMany) {
                record[name] = _this.loadHasManyRelation(base, attr, relation);
                return record;
            }
            if (attr.type === AttrTypes_1.default.HasManyBy) {
                record[name] = _this.loadHasManyByRelation(base, attr, relation);
                return record;
            }
            return record;
        }, _record);
    };
    /**
     * Load the has one relationship for the record.
     */
    Repo.prototype.loadHasOneRelation = function (record, attr, relation) {
        var entity = this.resolveRelation(attr).entity;
        var field = attr.foreignKey;
        var query = this.self().query(this.state, entity, false).where(field, record.id);
        this.addConstraint(query, relation);
        return query.first();
    };
    /**
     * Load the belongs to relationship for the record.
     */
    Repo.prototype.loadBelongsToRelation = function (record, attr, relation) {
        var entity = this.resolveRelation(attr).entity;
        var id = record[attr.foreignKey];
        var query = this.self().query(this.state, entity, false);
        this.addConstraint(query, relation);
        return query.first(id);
    };
    /**
     * Load the has many relationship for the record.
     */
    Repo.prototype.loadHasManyRelation = function (record, attr, relation) {
        var entity = this.resolveRelation(attr).entity;
        var field = attr.foreignKey;
        var query = this.self().query(this.state, entity, false).where(field, record.id);
        this.addConstraint(query, relation);
        return query.get();
    };
    /**
     * Load the has many by relationship for the record.
     */
    Repo.prototype.loadHasManyByRelation = function (record, attr, relation) {
        var _this = this;
        var entity = this.resolveRelation(attr).entity;
        var field = attr.foreignKey;
        return record[field].map(function (id) {
            var query = _this.self().query(_this.state, entity, false).where(attr.otherKey, id);
            _this.addConstraint(query, relation);
            return query.first();
        });
    };
    /**
     * Resolve relation out of the container.
     */
    Repo.prototype.resolveRelation = function (attr) {
        if (!_.isString(attr.model)) {
            return attr.model;
        }
        return this.model(name);
    };
    /**
     * Check if the given record has given relationship.
     */
    Repo.prototype.hasRelation = function (record, name, constraint, count) {
        if (constraint === void 0) { constraint = null; }
        var _constraint = constraint;
        if (typeof constraint === 'number') {
            _constraint = function (query) { return query.count() === constraint; };
        }
        else if (constraint === '>' && typeof count === 'number') {
            _constraint = function (query) { return query.count() > count; };
        }
        else if (constraint === '>=' && typeof count === 'number') {
            _constraint = function (query) { return query.count() >= count; };
        }
        else if (constraint === '<' && typeof count === 'number') {
            _constraint = function (query) { return query.count() < count; };
        }
        else if (constraint === '<=' && typeof count === 'number') {
            _constraint = function (query) { return query.count() <= count; };
        }
        var data = this.loadRelations(record, [{ name: name, constraint: _constraint }]);
        return !_.isEmpty(data[name]);
    };
    /**
     * Add constraint to the query.
     */
    Repo.prototype.addConstraint = function (query, relation) {
        var relations = relation.name.split('.');
        if (relations.length !== 1) {
            relations.shift();
            query.with(relations.join('.'));
            return;
        }
        var result = relation.constraint && relation.constraint(query);
        if (typeof result === 'boolean') {
            query.where(function () { return result; });
        }
    };
    /**
     * Normalize the given data by given model.
     */
    Repo.prototype.normalize = function (data) {
        return this.model(this.name).normalize(data);
    };
    return Repo;
}());
exports.default = Repo;
//# sourceMappingURL=Repo.js.map